"""Batch commands to be started with Fabric.

Imported functions should be prepended with underscore to hide them from
``fabric --list``.

Git branches should be named as ``staging`` for STAGING server and
``production`` for PRODUCTION one.

Required settings should be placed in project settings::

    APPS_TO_TEST = (
        'sms',
        'testing',
        'timezone',
        'translations',
    )

    # Project repositories.
    PROD_REPO = 'git@github.com:user/mysite.git'
    DEV_REPO = 'git@github.com:webengineer/mysite.git'
    PULL_REPO = DEV_REPO    # PROD_REPO for gatekeeper.

    # Logging setup.
    LOGS = {
        'common': 'var/messages.log',
        'payment': 'var/payments.log',
        'exception': 'var/errors.log',
        'sms': 'var/sms.log',
    }

    # Staging and production servers.
    STAGING = {
            'hostname': 'staging.dot.com',
            'ec2_instance': 'i-d86a73b0',
            'ip': '204.236.215.216',
            'proj_dir': '/var/www/ott/mysite',
            'virtualenv_dir': 've',
    }
    PRODUCTION = {
            'hostname': 'translation.dot.com',
            'ec2_instance': 'i-67ae4b0c',
            'ip': '184.73.37.110',
            'proj_dir': '/var/www/ott/mysite',
            'virtualenv_dir': 've',
    }
    VIRTUALENV_DIR = 've'

    # Project repository access key.
    SSH_KEYFILE = '/home/user/.ssh/id_user'

    # Should be same on all instances.
    MAINTENANCE_FILE = 'var/maintenance.lck'

Optional settings::

    # Managing Websitepulse alerts.
    WB_PULSE_USERNAME = 'odeskps'
    WB_PULSE_API_KEY = 'OVERRIDE ME IN LOCAL SETTINGS'
    WB_PULSE_LOGIN_TARGET = '73812'
    WB_PULSE_HEALTHCHECK_TARGET = '79432'

    SUDO_EXPORT = 'HOME=/home/odeskps/ PYTHON_EGG_CACHE=/tmp'
"""

from os.path import exists as _exists
from re import search as _search
from time import sleep as _sleep
import urllib2 as _urllib2, urllib as _urllib

from boto.ec2.connection import EC2Connection as _EC2Connection
from fabric.api import abort, cd, env, hosts, local, prompt, run, settings
from fabric.contrib.console import confirm

# TODO Provide in project settings.
from settings import (AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY,
        APPS_TO_TEST, PROD_REPO, DEV_REPO, PULL_REPO, LOGS,
        MAINTENANCE_FILE, PRODUCTION, SSH_KEYFILE, STAGING,
        VIRTUALENV_DIR)

try:
    from settings import (WB_PULSE_USERNAME, WB_PULSE_API_KEY,
            WB_PULSE_LOGIN_TARGET, WB_PULSE_HEALTHCHECK_TARGET)
except ImportError:
    NO_WB_PULSE_CONFIG = True
else:
    NO_WB_PULSE_CONFIG = False

try:
    from settings import SUDO_EXPORT
except:
    SUDO_EXPORT = ''


WB_API_URL='http://api.websitepulse.com/textserver.php'

env.key_filename = SSH_KEYFILE
env.user = 'odeskps'

SUDO_VIRTUALENV = 'source %s/bin/activate && %s ' % (VIRTUALENV_DIR, SUDO_EXPORT)
MANAGE = './manage.py '


def _check_git_branch_name(name):
    with settings(warn_only=True):
        res = local('git check-ref-format --branch %s' % name)
        if res.failed:
            raise Exception('"%s" is not a valid branch name' % name)
        else:
            return name


def _get_current_branch():
    """Returns name of current git branch.

    Creates new branch if not currently on any branch."""
    branches = local('git branch')
    branch = _search(r'\* (?P<branch>\(no branch\)|\S+)', branches).group(1)
    if branch == '(no branch)':
        local('git branch -r', capture=False)
        branch = prompt('Please specify new branch name:',
                        validate=_check_git_branch_name)
        local('git checkout -b %s' % branch)
    return branch


def pull_updates():
    """Pulls updates from remote staging and production branches."""
    local('git fetch %s staging:staging' % PULL_REPO)
    local('git checkout production')
    local('git pull %s production' % PULL_REPO)


def _disable_websitepulse_alerts(suspend=True):
    if NO_WB_PULSE_CONFIG: return
    data = {'username': WB_PULSE_USERNAME,
            'key': WB_PULSE_API_KEY,
            'action': 'suspend' if suspend else 'activate',
            'method': 'UpdateTargetStatus',
            'format': 'xml',
            'target': ','.join([WB_PULSE_LOGIN_TARGET,
                                WB_PULSE_HEALTHCHECK_TARGET])}
    print "%s monitoring for all targets" % (
            "Suspending" if suspend else "Activating")
    response = _urllib2.urlopen(_urllib2.Request(WB_API_URL,
                                               _urllib.urlencode(data))).read()


def _enable_websitepulse_alerts():
    _disable_websitepulse_alerts(suspend=False)


def prepare_deploy():
    """Run tests and commit staged chages."""
    feature = _get_current_branch()
    with settings(warn_only=True):
        res = local('git commit', capture=False)
        if res.failed and not confirm('Continue anyway?'):
            abort("Aborting at user request.")
    pull_updates()
    local('git checkout %s' % feature)
    with settings(warn_only=True):
        merge = local('git merge production')
    if merge.failed:
        local('git mergetool', capture=False)
        local('git commit', capture=False)
    local('python bootstrap.py')
    # Run tests.
    local(MANAGE + 'clean_pyc')
    local(MANAGE + 'test --settings=test_settings '
          + ' '.join(APPS_TO_TEST), capture=False)
    local(MANAGE + 'dryrun', capture=False)
    local(MANAGE + 'syncdb --noinput', capture = False)
    local(MANAGE + 'migrate --noinput', capture = False)
    if confirm('Do you like to run healthcheck?', default=False):
        local(MANAGE + 'healthcheck', capture=False)


@hosts(PRODUCTION['hostname'])
def backup_production():
    """Create backup dump and transfer to the STAGING server."""
    with cd(PRODUCTION['proj_dir']):
        run(SUDO_VIRTUALENV + 'bin/backup_db.sh', pty=True)
        run(SUDO_VIRTUALENV + 'bin/xfer_backups.sh', pty=True)


def _run_pull():
    """Update remote repo.

    Should be started within project directory."""
    if env['host_string'] == STAGING['hostname']:
        run('git checkout staging', pty=True)
    run('git pull', pty=True)
    # Setup virtualenv.
    run('python bootstrap.py')
    for log_file in LOGS.values():
        if not _exists(log_file):
            run('touch %s' % log_file, pty=True)
            run('chmod a+w %s' % log_file, pty=True)


def _run_migrate_restart():
    """Migrates DB and restarts HTTP server.

    Should be started within project directory."""
    run(MANAGE + 'syncdb --noinput', pty=True)
    run(MANAGE + 'migrate --noinput', pty=True)
    # Apache's mod_wsgi is used in daemon mode, so it's enough
    # to touch wsgi script to restart mod_wsgi processes without
    # restarting whole Apache.
    run('touch django.wsgi', pty=True)


def _run_tests():
    """Run tests.

    Should be started within project directory."""
    run(MANAGE + 'clean_pyc', pty=True)
    run(MANAGE + 'test --settings=test_settings ' + ' '.join(APPS_TO_TEST),
         pty=True)
    run(MANAGE + 'dryrun', pty=True)


def _run_healthcheck():
    """Restart remote server and test health.

    Should be started within project directory."""
    if env['host_string'] in (STAGING['hostname'], PRODUCTION['hostname']):
        host = 'http://%s/' % env['host_string']
        return run(MANAGE + 'healthcheck --realserver="%s"' % host, pty=True)
    else:
        return run(MANAGE + 'healthcheck', pty=True)


@hosts(PRODUCTION['hostname'])
def maintain_production(restore=False):
    """Enter maintenance mode on PRODUCTION."""
    with cd(PRODUCTION['proj_dir']):
        if not restore:
            run('touch %s' % MAINTENANCE_FILE, pty=True)
            _disable_websitepulse_alerts()
        else:
            run('rm %s' % MAINTENANCE_FILE, pty=True)
            _enable_websitepulse_alerts()

        # Apache's mod_wsgi is used in daemon mode, so it's enough
        # to touch wsgi script to restart mod_wsgi processes without
        # restarting whole Apache.
        run('touch django.wsgi', pty=True)


@hosts(PRODUCTION['hostname'])
def restore_production():
    """Exit from maintenance mode on PRODUCTION."""
    maintain_production(restore=True)


@hosts(STAGING['hostname'])
def deploy_staging():
    """Merges current branch into STAGING."""
    feature = _get_current_branch()
    local('git checkout staging')
    with settings(warn_only=True):
        merge = local('git merge --no-ff ' + feature)
    if merge.failed:
        local('git mergetool', capture=False)
    with settings(warn_only=True):
        local('git commit', capture=False)
    local(' '.join(['git push -f', DEV_REPO, 'staging']))
    local(' '.join(['git push', PROD_REPO, 'staging']))
    with cd(STAGING['proj_dir']):
        _run_pull()
        _run_tests()
        _run_migrate_restart()
        with settings(warn_only=True):
            res = _run_healthcheck()
            if res.failed and not confirm('Continue anyway?'):
                abort('Aborting at user request.')


@hosts(PRODUCTION['hostname'])
def deploy_production(maintain=False):
    """Deploy staging branch into PRODUCTION."""
    maintain = confirm('Do you like to ban users with maintenance mode?',
                       default=maintain)
    local('git checkout production')
    with settings(warn_only=True):
        merge = local('git merge staging')
    if merge.failed:
        local('git mergetool', capture=False)
        local('git commit', capture=False)
    local(' '.join(['git push -f', DEV_REPO, 'production']))
    local(' '.join(['git push', PROD_REPO, 'production']))
    backup_production()
    with cd(PRODUCTION['proj_dir']):
        if maintain:
            maintain_production()
        else:
            _disable_websitepulse_alerts()
        _run_pull()
        _run_migrate_restart()
        run('sudo -u www-data crontab -u www-data crontab', pty=True)
        _run_healthcheck()
        if maintain:
            restore_production()
        else:
            _enable_websitepulse_alerts()


def deploy(maintain=False):
    """Deploy staged updates of current branch."""
    start_staging()
    prepare_deploy()
    with settings(host_string=STAGING['hostname']):
        deploy_staging()
    if not confirm('STAGING updated, do you like to proceed with PRODUCTION?'):
        abort('Aborting at user request.')
    with settings(host_string=PRODUCTION['hostname']):
        deploy_production(maintain)
    if confirm('Do you like to shutdown STAGING server?'):
        stop_staging()


def stop_staging():
    """Shut down STAGING server to prevent environment pollution."""
    print 'Logging into Amazon AWS....'
    conn = _EC2Connection(AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
    print 'Stopping ' + STAGING['hostname'] + ' server:'
    print conn.stop_instances(STAGING['ec2_instance'])


def start_staging():
    """Start STAGING server."""
    print 'Logging into Amazon AWS....'
    conn = _EC2Connection(AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
    print 'Starting ' + STAGING['hostname'] + ' server:'
    print conn.start_instances(STAGING['ec2_instance'])
    _sleep(10) # seconds. Otherwise will get "The instance is not in the
              # 'running' state" error.
    print 'Associate Elastic IP with ' + STAGING['hostname'] + ' server:'
    print conn.associate_address(STAGING['ec2_instance'], STAGING['ip'])
    print '(Note that it might take a couple of minutes for the Elastic IP to work....)'




from fabric.api import env, settings, sudo, abort, put, task
from os.path import isfile as _isfile

from django_fabfile.utils import config, get_inst_by_id


try:
    preconfigured_user = config.get('DEFAULT', 'USERNAME')
except:
    pass    # Expecting user to be provided as `-u` option.
else:
    FABRIC_DEFAULT_USER = 'user'    # XXX `-u user` will be overridden.
    if env['user'] == FABRIC_DEFAULT_USER:  # Not provided as `-u` option.
        env.update({'user': preconfigured_user})

env.update({'disable_known_hosts': True})


def _sudo(cmd):
    """ Shows output of cmd and allows interaction """
    sudo(cmd, shell=False, pty=True)


def _create_account(username, region, instance_ids, passwordless, sudo):
    if not _isfile(username + '.pub'):
        abort("%s.pub does not exist" % username)
    env.ssh_key = username + '.pub'
    env.username = username     # Own attribute for string formatting.
    if passwordless:
        _sudo('adduser --disabled-password %(username)s' % env)
        if sudo:
            _sudo('sed -i "s/# %sudo ALL=NOPASSWD: ALL/'
                            '%sudo ALL=NOPASSWD: ALL/" /etc/sudoers')
            for group in ['sudo']:
                with settings(group=group):
                    _sudo('adduser %(username)s %(group)s' % env)
    else:
        _sudo('adduser %(username)s' % env)
        if sudo:
            for group in ['adm', 'admin', 'staff']:
                with settings(group=group):
                    _sudo('adduser %(username)s %(group)s' % env)
    _sudo('mkdir -p /home/%(username)s/.ssh' % env)
    _sudo('touch /home/%(username)s/.ssh/authorized_keys' % env)
    _sudo('chown -R %(username)s: /home/%(username)s/.ssh' % env)
    _sudo('chmod 700 /home/%(username)s/.ssh' % env)
    put(env.ssh_key, '/home/%(username)s/.ssh/authorized_keys'
                                      % env, use_sudo=True)
    _sudo('chown -R %(username)s: /home/%(username)s/.'
                                    'ssh/authorized_keys' % env)
    _sudo('chmod 600 /home/%(username)s/.ssh/authorized_keys' % env)


@task
def deluser(name, region=None, instance_ids=None):
    """
    Removes user <name> with deluser from "host1;host2" list in <region>
    If region and instance_ids not set - script takes hosts amd key values
    from command line (-H and -i).
    """
    if instance_ids and region:
        instances_ids = list(unicode(instance_ids).split(';'))
        for inst in instances_ids:
            if inst:
                _instance = get_inst_by_id(region, inst)
                if not env.key_filename:
                    key_filename = config.get(_instance.region.name,
                                                      'KEY_FILENAME')
                    env.update({'key_filename': key_filename,
                                                  'warn_only': True})
                env.update({'host_string': _instance.public_dns_name})
                env.username = name
                _sudo('deluser %(username)s' % env)
    else:
        env.update({'warn_only': True})
        env.username = name
        _sudo('deluser %(username)s' % env)


@task
def adduser(username, region=None, instance_ids=None,
                                passwordless=None, sudo=None):
    """
    creates new <username> with public SSH key on "host1;host2" list in
    <region>. If you want to create passwordless account - set any value to
    <passwrdless> variable, if you want sudo rights - set any value to <sudo>.
    File with public key must be in same directory.
    If region and instance_ids not set - script takes hosts amd key values
    from command line (-H and -i).
    Usage:
    1. WIthout aws api keys and config present:
    :<username>,<passwordless=1>,<sudo=1> - in this case you have to specify
    hosts list in -H and your own account in -u fabric parameters.
    2. With aws api keys and config entries:
    :<username>,<region>,"instance1;instance2",<passwordless>,<sudo>
    Extracts IP's from instance description.
    """
    if instance_ids and region:
        instances_ids = list(unicode(instance_ids).split(';'))
        for inst in instances_ids:
            if inst:
                _instance = get_inst_by_id(region, inst)
                if not env.key_filename:
                    key_filename = config.get(_instance.region.name,
                                                      'KEY_FILENAME')
                    env.update({'key_filename': key_filename})
                env.update({'host_string': _instance.public_dns_name})
                _create_account(username, region, instance_ids, passwordless,
                                                                    sudo)
    else:
        _create_account(username, region, instance_ids, passwordless, sudo)



from collections import defaultdict
from contextlib import contextmanager
from itertools import chain, groupby
from hashlib import sha256
import logging
from operator import attrgetter
from warnings import warn

from boto.exception import EC2ResponseError
from fabric.api import task

from django_fabfile.utils import config, get_region_conn, timestamp


logger = logging.getLogger(__name__)


INST_SPECIFIC_SG_PREFIX = 'Created on '


def new_security_group(region, name=None, description=None):
    """Create Security Groups with SSH access."""
    s_g = get_region_conn(region.name).create_security_group(
        name or INST_SPECIFIC_SG_PREFIX + timestamp(),
        description or 'Created for using with specific instance')
    s_g.authorize('tcp', 22, 22, '0.0.0.0/0')
    return s_g


@task
def cleanup_security_groups(delete=False):
    """
    Delete unused AWS Security Groups.

    :type delete: boolean
    :param delete: notify only (i.e. False) by default.

    If security group with the same name is used at least in one region,
    it is treated as used.
    """
    groups = defaultdict(lambda: {})
    used_groups = set(['default',
                       config.get('DEFAULT', 'HTTPS_SECURITY_GROUP')])
    regions = get_region_conn().get_all_regions()
    for reg in regions:
        for s_g in get_region_conn(reg.name).get_all_security_groups():
            groups[s_g.name][reg] = s_g
            if s_g.instances():     # Security Group is used by instance.
                used_groups.add(s_g.name)
            for rule in s_g.rules:
                for grant in rule.grants:
                    if grant.name and grant.owner_id == s_g.owner_id:
                        used_groups.add(grant.name)     # SG is used by group.
    for grp in used_groups:
        del groups[grp]

    for grp in sorted(groups):
        if delete:
            for reg in groups[grp]:
                s_g = groups[grp][reg]
                logger.info('Deleting {0} in {1}'.format(s_g, reg))
                s_g.delete()
        else:
            msg = '"SecurityGroup:{grp}" should be removed from {regs}'
            logger.info(msg.format(grp=grp, regs=groups[grp].keys()))


def regroup_rules(security_group):
    grouped_rules = defaultdict(lambda: [])
    for rule in security_group.rules:
        ports = rule.ip_protocol, rule.from_port, rule.to_port
        for grant in rule.grants:
            grouped_rules[ports].append(grant)
    for rule in grouped_rules:  # Ordering for hashing.
        grouped_rules[rule] = tuple(sorted(grouped_rules[rule], key=str))
    return grouped_rules


def sync_rules(src_grp, dst_grp=None, dst_region=None):
    """
    Copy Security Group rules.

    Works across regions as well. The sole exception that won't be
    synced is granted groups, owned by another user - such groups can't
    be copied recursively.
    """
    assert bool(dst_grp) ^ bool(dst_region), ('Only dst_grp or dst_region '
                                              'should be provided')
    if dst_region:
        dst_grp = new_security_group(dst_region, src_grp.name,
                                                 src_grp.description)

    def is_group_in(region, group_name):
        try:
            get_region_conn(region.name).get_all_security_groups([group_name])
        except EC2ResponseError:
            return False
        else:
            return True

    src_rules = regroup_rules(src_grp)
    # Assure granted group represented in destination region.
    src_grants = chain(*src_rules.values())
    for grant in dict((grant.name, grant) for grant in src_grants).values():
        if (grant.name and grant.owner_id == src_grp.owner_id and
                not is_group_in(dst_grp.region, grant.name)):
            src_conn = get_region_conn(src_grp.region.name)
            grant_grp = src_conn.get_all_security_groups([grant.name])[0]
            sync_rules(grant_grp, dst_region=dst_grp.region)
    dst_rules = regroup_rules(dst_grp)
    # Remove rules absent in src_grp.
    for ports in set(dst_rules.keys()) - set(src_rules.keys()):
        for grant in dst_rules[ports]:
            args = ports + ((None, grant) if grant.name else (grant, None))
            dst_grp.revoke(*args)
    # Add rules absent in dst_grp.
    for ports in set(src_rules.keys()) - set(dst_rules.keys()):
        for grant in src_rules[ports]:
            if grant.name and not is_group_in(dst_grp.region, grant.name):
                continue    # Absent other's granted group.
            args = ports + ((None, grant) if grant.name else (grant, None))
            dst_grp.authorize(*args)
    # Refresh `dst_rules` from updated `dst_grp`.
    dst_rules = regroup_rules(dst_grp)

    @contextmanager
    def patch_grouporcidr():
        """XXX Patching `boto.ec2.securitygroup.GroupOrCIDR` cmp and hash."""
        from boto.ec2.securitygroup import GroupOrCIDR
        original_cmp = getattr(GroupOrCIDR, '__cmp__', None)
        GroupOrCIDR.__cmp__ = lambda self, other: cmp(str(self), str(other))
        original_hash = GroupOrCIDR.__hash__
        GroupOrCIDR.__hash__ = lambda self: hash(str(self))
        try:
            yield
        finally:
            if original_cmp:
                GroupOrCIDR.__cmp__ = original_cmp
            else:
                del GroupOrCIDR.__cmp__
            GroupOrCIDR.__hash__ = original_hash

    # Sync grants in common rules.
    with patch_grouporcidr():
        for ports in src_rules:
            # Remove grants absent in src_grp rules.
            for grant in set(dst_rules[ports]) - set(src_rules[ports]):
                args = ports + ((None, grant) if grant.name else (grant, None))
                dst_grp.revoke(*args)
            # Add grants absent in dst_grp rules.
            for grant in set(src_rules[ports]) - set(dst_rules[ports]):
                if grant.name and not is_group_in(dst_grp.region, grant.name):
                    continue    # Absent other's granted group.
                args = ports + ((None, grant) if grant.name else (grant, None))
                dst_grp.authorize(*args)


@task
def sync_rules_by_id(src_reg_name, src_grp_id, dst_reg_name, dst_grp_id):
    """Update Security Group rules from other Security Group.

    Works across regions as well. The sole exception is granted groups,
    owned by another user - such groups can't be copied.

    :param src_reg_name: region name
    :type src_reg_name: str
    :param src_grp_id: group ID
    :type src_grp_id: str
    :param dst_reg_name: region name
    :type dst_reg_name: str
    :param dst_grp_id: group ID
    :type dst_grp_id: str"""
    src_grp = get_region_conn(src_reg_name).get_all_security_groups(
        filters={'group-id': src_grp_id})[0]
    dst_grp = get_region_conn(dst_reg_name).get_all_security_groups(
        filters={'group-id': dst_grp_id})[0]
    sync_rules(src_grp, dst_grp)


@task
def replicate_security_groups(filters=None):
    """
    Replicate updates of Security Groups among regions.

    :param filters: restrict replication to subset of Security Groups,
        see available options at
        http://docs.amazonwebservices.com/AWSEC2/latest/APIReference/ApiReference-query-DescribeSecurityGroups.html.
        Not available while running as Fabric task because it should be
        of `dict` type.
    :type filters: dict


    Per-instance Security Groups without additional rules won't be
    replicated.

    Raises warnings about synchronization issues that requires manual
    resolution.
    """
    HASH, TIMESTAMP = 'Hash', 'Version'     # Tag names.

    def get_hash(s_g):
        """
        Return unique hash for Security Group rules.

        Granted Security Groups will be respected identical if them
        belongs to identical owner and identically named irrespectively
        to region.
        """
        return sha256(str(regroup_rules(s_g).items())).hexdigest()

    def was_updated(s_g):
        """Returns True if Security Group was modified or just created."""
        return HASH not in s_g.tags or get_hash(s_g) != s_g.tags[HASH]

    regions = get_region_conn().get_all_regions()
    blank_group = new_security_group(regions[0])
    security_groups = []
    for reg in regions:
        for s_g in get_region_conn(reg.name).get_all_security_groups(
                filters=filters):
            security_groups.append(s_g)
    name = attrgetter('name')
    grp_by_name = groupby(sorted(security_groups, key=name), key=name)
    for name, grp_in_regions in grp_by_name:
        grp_in_regions = list(grp_in_regions)
        versions = set(get_hash(s_g) for s_g in grp_in_regions)
        old_vers = [s_g for s_g in grp_in_regions if not was_updated(s_g)]
        if len(set(s_g.tags[HASH] for s_g in old_vers)) > 1:
            warn('Old versions of {0} should be synced manually'.format(name))
            continue
        if len(versions) == 2 and old_vers:  # Update olds to new version.
            new = [grp for grp in grp_in_regions if was_updated(grp)][0]
            for prev in old_vers:
                sync_rules(new, prev)
        elif not len(versions) == 1:
            warn('More than 1 new versions of {0} found. Synchronization '
                 'can\'t be applied.'.format(name))
            continue
        # Clone to all regions if not yet cloned.
        if (len(grp_in_regions) < len(regions) and
            not (name.startswith(INST_SPECIFIC_SG_PREFIX) and
                 get_hash(grp_in_regions[0]) == get_hash(blank_group))):
            s_g_regions = set(s_g.region.name for s_g in grp_in_regions)
            for reg_name in set(reg.name for reg in regions) - s_g_regions:
                region = get_region_conn(reg_name).region
                sync_rules(grp_in_regions[0], dst_region=region)
        # Update tags.
        mark = timestamp()
        for s_g in grp_in_regions:
            s_g.add_tag(HASH, get_hash(s_g))
            s_g.add_tag(TIMESTAMP, mark)
    blank_group.delete()



"""Check :doc:`README` or :class:`django_fabfile.utils.Config` docstring
for setup instructions."""

from contextlib import contextmanager
import logging
import os
import re
from datetime import timedelta, datetime
from contextlib import nested
from itertools import groupby
from json import dumps

from boto.exception import EC2ResponseError
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from dateutil.tz import tzutc
from fabric.api import env, local, put, settings, sudo, task, run
from fabric.contrib.files import append

from django_fabfile.instances import (attach_snapshot, create_temp_inst,
                                      get_avail_dev, get_vol_dev, mount_volume)
from django_fabfile.utils import (
    StateNotChangedError, add_tags, config, config_temp_ssh,
    get_descr_attr, get_inst_by_id, get_region_conn, get_snap_device,
    get_snap_time, get_snap_vol, timestamp, wait_for, wait_for_sudo)


USERNAME = config.get('DEFAULT', 'USERNAME')
env.update({'user': USERNAME, 'disable_known_hosts': True})

logger = logging.getLogger(__name__)


DEFAULT_TAG_NAME = config.get('DEFAULT', 'TAG_NAME')
DEFAULT_TAG_VALUE = config.get('DEFAULT', 'TAG_VALUE')
DESCRIPTION_TAG = 'Description'
SNAP_STATUSES = ['pending', 'completed']    # All but "error".
VOL_STATUSES = ['creating', 'available', 'in-use']
DETACH_TIME = config.getint('DEFAULT', 'MINUTES_FOR_DETACH') * 60
SNAP_TIME = config.getint('DEFAULT', 'MINUTES_FOR_SNAP') * 60
REPLICATION_SPEED = config.getfloat('DEFAULT', 'REPLICATION_SPEED')


class ReplicationCollisionError(Exception):
    pass


def create_snapshot(vol, description='', tags=None, synchronously=True,
                    consistent=False):
    """Return new snapshot for the volume.

    vol
        volume to snapshot;
    synchronously
        wait for successful completion;
    description
        description for snapshot. Will be compiled from instnace
        parameters by default;
    tags
        tags to be added to snapshot. Will be cloned from volume and from
        instance by default.
    consistent
        if consistent True, script will try to freeze fs mountpoint and create
        snapshot while it's freezed with all buffers dumped to disk.
    """
    if vol.attach_data:
        inst = get_inst_by_id(vol.region.name, vol.attach_data.instance_id)
    else:
        inst = None
    if not description and inst:
        description = dumps({
            'Volume': vol.id,
            'Region': vol.region.name,
            'Device': vol.attach_data.device,
            'Instance': inst.id,
            'Type': inst.instance_type,
            'Arch': inst.architecture,
            'Root_dev_name': inst.root_device_name,
            'Time': timestamp(),
            })

    def freeze_volume():
        key_filename = config.get(inst.region.name, 'KEY_FILENAME')
        try:
            _user = config.get('SYNC', 'USERNAME')
        except:
            _user = USERNAME
        with settings(host_string=inst.public_dns_name,
                      key_filename=key_filename, user=_user):
            wait_for_sudo('sync', shell=False)
            run('for i in {1..20}; do sudo sync; sleep 1; done &')

    def initiate_snapshot():
        if consistent:
            if inst.state == 'running':
                try:
                    freeze_volume()
                except:
                    logger.info('FS NOT FREEZED! '
                                'Do you have access to this server?')
        snapshot = vol.create_snapshot(description)
        if tags:
            add_tags(snapshot, tags)
        else:
            add_tags(snapshot, vol.tags)
            if inst:
                add_tags(snapshot, inst.tags)
        logger.info('{0} started from {1} in {0.region}'.format(snapshot, vol))
        return snapshot

    if synchronously:
        while True:     # Iterate unless success and delete failed snapshots.
            snapshot = initiate_snapshot()
            try:
                wait_for(snapshot, '100%', limit=SNAP_TIME)
                assert snapshot.status == 'completed', (
                    'completed with wrong status {0}'.format(snapshot.status))
            except (StateNotChangedError, AssertionError) as err:
                logger.error(str(err) + ' - deleting')
                snapshot.delete()
            else:
                break
    else:
        snapshot = initiate_snapshot()
    return snapshot


@task
def backup_instance(region_name, instance_id=None, instance=None,
                    synchronously=False, consistent=False):
    """
    Return list of created snapshots for specified instance.

    region_name
        instance location;
    instance, instance_id
        either `instance_id` or `instance` argument should be specified;
    synchronously
        wait for successful completion. False by default.
    consistent
        if True, then FS mountpoint will be frozen before snapshotting.
        False by default.
    """
    assert bool(instance_id) ^ bool(instance), ('Either instance_id or '
        'instance should be specified')
    conn = get_region_conn(region_name)
    if instance_id:
        instance = get_inst_by_id(conn.region.name, instance_id)
    snapshots = []
    for dev in instance.block_device_mapping:
        vol_id = instance.block_device_mapping[dev].volume_id
        vol = conn.get_all_volumes([vol_id])[0]
        snapshots.append(create_snapshot(vol, synchronously=synchronously,
                         consistent=consistent))
    return snapshots


@task
def backup_instances_by_tag(
        region_name=None, tag_name=DEFAULT_TAG_NAME,
        tag_value=DEFAULT_TAG_VALUE, synchronously=False, consistent=False):
    """Creates backup for all instances with given tag in region.

    region_name
        will be applied across all regions by default;
    tag_name, tag_value
        will be fetched from config by default;
    synchronously
        will be accomplished with assuring successful result. False by
        default;
    consistent
        if True, then FS mountpoint will be frozen before snapshotting.

    .. note:: when ``create_ami`` task compiles AMI from several
              snapshots it restricts snapshot start_time difference with
              10 minutes interval at most. Snapshot completion may take
              much more time and due to this only asynchronously
              generated snapshots will be assembled assurely."""
    if region_name:
        regions = [get_region_conn(region_name).region]
    else:
        regions = get_region_conn().get_all_regions()
    for reg in regions:
        conn = get_region_conn(reg.name)
        filters = {'resource-type': 'instance', 'key': tag_name,
                   'tag-value': tag_value}
        for tag in conn.get_all_tags(filters=filters):
            backup_instance(reg.name, instance_id=tag.res_id,
                            synchronously=synchronously, consistent=consistent)


def _trim_snapshots(region, dry_run=False):

    """Delete snapshots back in time in logarithmic manner.

    dry_run
        just print snapshot to be deleted.

    Modified version of the `boto.ec2.connection.trim_snapshots
    <http://pypi.python.org/pypi/boto/2.0>_`. Licensed under MIT license
    by Mitch Garnaat, 2011."""
    hourly_backups = config.getint('purge_backups', 'HOURLY_BACKUPS')
    daily_backups = config.getint('purge_backups', 'DAILY_BACKUPS')
    weekly_backups = config.getint('purge_backups', 'WEEKLY_BACKUPS')
    monthly_backups = config.getint('purge_backups', 'MONTHLY_BACKUPS')
    quarterly_backups = config.getint('purge_backups', 'QUARTERLY_BACKUPS')
    yearly_backups = config.getint('purge_backups', 'YEARLY_BACKUPS')

    # work with UTC time, which is what the snapshot start time is reported in
    now = datetime.utcnow()
    last_hour = datetime(now.year, now.month, now.day, now.hour)
    last_midnight = datetime(now.year, now.month, now.day)
    last_sunday = datetime(now.year, now.month,
          now.day) - timedelta(days=(now.weekday() + 1) % 7)
    last_month = datetime.now() - relativedelta(months=1)
    last_year = datetime.now() - relativedelta(years=1)
    other_years = datetime.now() - relativedelta(years=2)
    start_of_month = datetime(now.year, now.month, 1)

    target_backup_times = []
    # there are no snapshots older than 1/1/2000
    oldest_snapshot_date = datetime(2000, 1, 1)

    for hour in range(0, hourly_backups):
        target_backup_times.append(last_hour - timedelta(hours=hour))

    for day in range(0, daily_backups):
        target_backup_times.append(last_midnight - timedelta(days=day))

    for week in range(0, weekly_backups):
        target_backup_times.append(last_sunday - timedelta(weeks=week))

    for month in range(0, monthly_backups):
        target_backup_times.append(last_month - relativedelta(months=month))

    for quart in range(0, quarterly_backups):
        target_backup_times.append(last_year - relativedelta(months=4 * quart))

    for year in range(0, yearly_backups):
        target_backup_times.append(other_years - relativedelta(years=year))

    one_day = timedelta(days=1)
    while start_of_month > oldest_snapshot_date:
        # append the start of the month to the list of snapshot dates to save:
        target_backup_times.append(start_of_month)
        # there's no timedelta setting for one month, so instead:
        # decrement the day by one,
        #so we go to the final day of the previous month...
        start_of_month -= one_day
        # ... and then go to the first day of that previous month:
        start_of_month = datetime(start_of_month.year,
                               start_of_month.month, 1)

    temp = []

    for t in target_backup_times:
        if temp.__contains__(t) == False:
            temp.append(t)

    target_backup_times = temp
    target_backup_times.reverse()  # make the oldest date first

    # get all the snapshots, sort them by date and time,
    #and organize them into one array for each volume:
    conn = get_region_conn(region.name)
    all_snapshots = conn.get_all_snapshots(owner='self')
    # oldest first
    all_snapshots.sort(cmp=lambda x, y: cmp(x.start_time, y.start_time))

    snaps_for_each_volume = {}
    for snap in all_snapshots:
        # the snapshot name and the volume name are the same.
        # The snapshot name is set from the volume
        # name at the time the snapshot is taken
        volume_name = get_snap_vol(snap)

        if volume_name:
            # only examine snapshots that have a volume name
            snaps_for_volume = snaps_for_each_volume.get(volume_name)

            if not snaps_for_volume:
                snaps_for_volume = []
                snaps_for_each_volume[volume_name] = snaps_for_volume
            snaps_for_volume.append(snap)

    # Do a running comparison of snapshot dates to desired time periods,
    # keeping the oldest snapshot in each
    # time period and deleting the rest:
    for volume_name in snaps_for_each_volume:
        snaps = snaps_for_each_volume[volume_name]
        snaps = snaps[:-1]
        # never delete the newest snapshot, so remove it from consideration

        time_period_num = 0
        snap_found_for_this_time_period = False
        for snap in snaps:
            check_this_snap = True

            while (check_this_snap and
                   time_period_num < target_backup_times.__len__()):

                if get_snap_time(snap) < target_backup_times[time_period_num]:
                    # the snap date is before the cutoff date.
                    # Figure out if it's the first snap in this
                    # date range and act accordingly
                    #(since both date the date ranges and the snapshots
                    # are sorted chronologically, we know this
                    #snapshot isn't in an earlier date range):
                    if snap_found_for_this_time_period:
                        if not snap.tags.get('preserve_snapshot'):
                            if dry_run:
                                logger.info('Dry-trimmed {0} {1} from {2}'
                                    .format(snap, snap.description,
                                    snap.start_time))
                            else:
                                # as long as the snapshot wasn't marked with
                                # the 'preserve_snapshot' tag, delete it:
                                try:
                                    conn.delete_snapshot(snap.id)
                                except EC2ResponseError as err:
                                    logger.exception(str(err))
                                else:
                                    logger.info('Trimmed {0} {1} from {2}'
                                        .format(snap, snap.description,
                                        snap.start_time))
                       # go on and look at the next snapshot,
                       # leaving the time period alone
                    else:
                        # this was the first snapshot found for this time
                        # period. Leave it alone and look at the next snapshot:
                        snap_found_for_this_time_period = True
                    check_this_snap = False
                else:
                    # the snap is after the cutoff date.
                    # Check it against the next cutoff date
                    time_period_num += 1
                    snap_found_for_this_time_period = False


@task
def delete_broken_snapshots():
    """Delete snapshots with status 'error'."""
    for region in get_region_conn().get_all_regions():
        conn = get_region_conn(region.name)
        filters = {'status': 'error'}
        snaps = conn.get_all_snapshots(owner='self', filters=filters)
        for snp in snaps:
            logger.info('Deleting broken {0}'.format(snp))
            snp.delete()


@task
def trim_snapshots(region_name=None, dry_run=False):
    """Delete old snapshots logarithmically back in time.

    region_name
        by default process all regions;
    dry_run
        boolean, only print info about old snapshots to be deleted."""
    delete_broken_snapshots()
    if region_name:
        regions = [get_region_conn(region_name).region]
    else:
        regions = get_region_conn().get_all_regions()
    for reg in regions:
        logger.info('Processing {0}'.format(reg))
        _trim_snapshots(reg, dry_run=dry_run)


@task
def rsync_mountpoints(src_inst, src_vol, src_mnt, dst_inst, dst_vol, dst_mnt,
                      encr=False):
    """Run `rsync` against mountpoints, copy disk label.

    :param src_inst: source instance;
    :param src_vol: source volume with label that will be copied to
                    dst_vol;
    :param src_mnt: root or directory hierarchy to replicate;
    :param dst_inst: destination instance;
    :param dst_vol: destination volume, that will be marked with label
                    from src_vol;
    :param dst_mnt: destination point where source hierarchy to place;
    :param encr: True if volume is encrypted;
    :type encr: bool."""
    src_key_filename = config.get(src_inst.region.name, 'KEY_FILENAME')
    dst_key_filename = config.get(dst_inst.region.name, 'KEY_FILENAME')
    with config_temp_ssh(dst_inst.connection) as key_file:
        with settings(host_string=dst_inst.public_dns_name,
                      key_filename=dst_key_filename):
            wait_for_sudo('cp /root/.ssh/authorized_keys '
                          '/root/.ssh/authorized_keys.bak')
            pub_key = local('ssh-keygen -y -f {0}'.format(key_file), True)
            append('/root/.ssh/authorized_keys', pub_key, use_sudo=True)
            if encr:
                sudo('screen -d -m sh -c "nc -l 60000 | gzip -dfc | '
                     'sudo dd of={0} bs=16M"'
                     .format(get_vol_dev(dst_vol)), pty=False)  # dirty magick
                dst_ip = sudo(
                    'curl http://169.254.169.254/latest/meta-data/public-ipv4')

        with settings(host_string=src_inst.public_dns_name,
                      key_filename=src_key_filename):
            put(key_file, '.ssh/', mirror_local_mode=True)
            dst_key_filename = os.path.split(key_file)[1]
            if encr:
                sudo('(dd if={0} bs=16M | gzip -cf --fast | nc -v {1} 60000)'
                     .format(get_vol_dev(src_vol), dst_ip))
            else:
                cmd = (
                    'rsync -e "ssh -i .ssh/{key_file} -o '
                    'StrictHostKeyChecking=no" -cahHAX --delete --inplace '
                    '--exclude /root/.bash_history '
                    '--exclude /home/*/.bash_history '
                    '--exclude /etc/ssh/moduli --exclude /etc/ssh/ssh_host_* '
                    '--exclude /etc/udev/rules.d/*persistent-net.rules '
                    '--exclude /var/lib/ec2/* --exclude=/mnt/* '
                    '--exclude=/proc/* --exclude=/tmp/* '
                    '{src_mnt}/ root@{rhost}:{dst_mnt}')
                wait_for_sudo(cmd.format(
                    rhost=dst_inst.public_dns_name, dst_mnt=dst_mnt,
                    key_file=dst_key_filename, src_mnt=src_mnt))
                label = sudo('e2label {0}'.format(get_vol_dev(src_vol)))
        with settings(host_string=dst_inst.public_dns_name,
                      key_filename=dst_key_filename):
            if not encr:
                sudo('e2label {0} {1}'.format(get_vol_dev(dst_vol), label))
            wait_for_sudo('mv /root/.ssh/authorized_keys.bak '
                          '/root/.ssh/authorized_keys')
            wait_for_sudo('sync', shell=False)
            wait_for_sudo('for i in {1..20}; do sync; sleep 1; done &')


def update_snap(src_vol, src_mnt, dst_vol, dst_mnt, encr, delete_old=False):

    """Update destination region from `src_vol`.

    Create new snapshot with same description and tags. Delete previous
    snapshot (if exists) of the same volume in destination region if
    ``delete_old`` is True."""

    src_inst = get_inst_by_id(src_vol.region.name,
                              src_vol.attach_data.instance_id)
    dst_inst = get_inst_by_id(dst_vol.region.name,
                              dst_vol.attach_data.instance_id)
    rsync_mountpoints(src_inst, src_vol, src_mnt, dst_inst, dst_vol, dst_mnt,
                     encr)
    src_snap = src_vol.connection.get_all_snapshots([src_vol.snapshot_id])[0]
    create_snapshot(dst_vol, description=src_snap.description,
                                    tags=src_snap.tags, synchronously=False)
    if delete_old and dst_vol.snapshot_id:
        old_snap = dst_vol.connection.get_all_snapshots(
            [dst_vol.snapshot_id])[0]
        logger.info('Deleting previous {0} in {1}'.format(old_snap,
                                                          dst_vol.region))
        old_snap.delete()


@contextmanager
def create_tmp_volume(region, size):
    """Format new filesystem."""
    with create_temp_inst(region) as inst:
        earmarking_tag = config.get(region.name, 'TAG_NAME')
        try:
            vol = get_region_conn(region.name).create_volume(size,
                                                             inst.placement)
            vol.add_tag(earmarking_tag, 'temporary')
            vol.attach(inst.id, get_avail_dev(inst))
            yield vol, mount_volume(vol, mkfs=True)
        finally:
            vol.detach(force=True)
            wait_for(vol, 'available', limit=DETACH_TIME)
            vol.delete()


def get_relevant_snapshots(
        conn, tag_name=DEFAULT_TAG_NAME, tag_value=DEFAULT_TAG_VALUE,
        native_only=True, filters={'status': SNAP_STATUSES}):
    """Returns snapshots with proper description."""
    if tag_name and tag_value:
        filters.update({'tag:{0}'.format(tag_name): tag_value})
    snaps = conn.get_all_snapshots(owner='self', filters=filters)
    is_described = lambda snap: get_snap_vol(snap) and get_snap_time(snap)
    snaps = [snp for snp in snaps if is_described(snp)]
    if native_only:
        is_native = lambda snp, reg: get_descr_attr(snp, 'Region') == reg.name
        snaps = [snp for snp in snaps if is_native(snp, conn.region)]
    return snaps


def get_replicas(descriptions, dst_conn):
    snaps = [snp for snp in dst_conn.get_all_snapshots(
        owner='self', filters={'status': SNAP_STATUSES,
                               'description': descriptions})]
    # Temporary volumes used by in-process replication.
    vols = [vol for vol in dst_conn.get_all_volumes(filters={
        'tag:{0}'.format(DESCRIPTION_TAG): descriptions,
        'status': VOL_STATUSES})]
    return snaps, vols


def get_oldest_replica(
        src_conn, dst_conn, amount=1, native_only=True,
        tag_name=DEFAULT_TAG_NAME, tag_value=DEFAULT_TAG_VALUE):
    """Return list with (one by default) not yet replicated snapshots."""
    snaps = get_relevant_snapshots(src_conn, tag_name, tag_value, native_only)
    if not snaps:
        return []
    # Separating out all but latest snapshots for every volume.
    latest_snaps = []
    snaps = sorted(snaps, key=get_snap_vol)
    for vol_id, vol_snaps in groupby(snaps, key=get_snap_vol):
        latest_snaps.append(sorted(vol_snaps, key=get_snap_time)[-1])
    # Seeking for latests replicas in dst region for every new snapshot.
    latest_descriptions = [snp.description for snp in latest_snaps]
    dst_snaps, dst_vols = get_replicas(latest_descriptions, dst_conn)
    snap_desc = [snp.description for snp in dst_snaps]
    vol_desc = [vol.tags[DESCRIPTION_TAG] for vol in dst_vols]
    # Seeking for snaps wihtout replicas.
    snaps_to_replicate = [snp for snp in latest_snaps if
        snp.description not in set(snap_desc + vol_desc)]
    return sorted(snaps_to_replicate, key=get_snap_time)[:amount]


@task
def rsync_snapshot(src_region_name, snapshot_id, dst_region_name,
                   src_inst=None, dst_inst=None, force=False):

    """Duplicate the snapshot into dst_region.

    src_region_name, dst_region_name
        Amazon region names. Allowed to be contracted, e.g.
        `ap-southeast-1` will be recognized in `ap-south` or even
        `ap-s`;
    snapshot_id
        snapshot to duplicate;
    src_inst, dst_inst
        will be used instead of creating new for temporary;
    force
        rsync snapshot even if newer version exist.

    You'll need to open port 60000 for encrypted instances replication."""
    src_conn = get_region_conn(src_region_name)
    src_snap = src_conn.get_all_snapshots([snapshot_id])[0]
    dst_conn = get_region_conn(dst_region_name)
    _src_device = get_snap_device(src_snap)
    _src_dev = re.match(r'^/dev/sda$', _src_device)  # check for encryption
    if _src_dev:
        encr = True
        logger.info('Found traces of encryption')
    else:
        encr = None

    info = 'Going to transmit {snap.volume_size} GiB {snap} {snap.description}'
    if src_snap.tags.get('Name'):
        info += ' of {name}'
    info += ' from {snap.region} to {dst}'
    logger.info(info.format(snap=src_snap, dst=dst_conn.region,
                            name=src_snap.tags.get('Name')))

    src_vol = get_snap_vol(src_snap)
    dst_snaps = get_relevant_snapshots(dst_conn, native_only=False)
    vol_snaps = [snp for snp in dst_snaps if get_snap_vol(snp) == src_vol]

    def sync_mountpoints(src_snap, src_vol, src_mnt, dst_vol, dst_mnt):
        # Marking temporary volume with snapshot's description.
        dst_vol.add_tag(DESCRIPTION_TAG, src_snap.description)
        snaps, vols = get_replicas(src_snap.description, dst_vol.connection)
        if not force and snaps:
            raise ReplicationCollisionError(
                'Stepping over {snap} - it\'s already replicated as {snaps} '
                'in {snaps[0].region}'.format(snap=src_snap, snaps=snaps))
        if not force and len(vols) > 1:
            timeout = src_snap.volume_size / REPLICATION_SPEED
            get_vol_time = lambda vol: parse(vol.create_time)

            def not_outdated(vol, now):
                age = now - get_vol_time(vol)
                return age.days * 24 * 60 * 60 + age.seconds < timeout

            now = datetime.utcnow().replace(tzinfo=tzutc())
            actual_vols = [vol for vol in vols if not_outdated(vol, now)]
            hunged_vols = set(vols) - set(actual_vols)
            if len(actual_vols) > 1:
                oldest = sorted(actual_vols, key=get_vol_time)[0]
                if dst_vol.id != oldest.id:
                    raise ReplicationCollisionError(
                        'Stepping over {snap} - it\'s already replicating to '
                        '{vol} in {vol.region}'.format(snap=src_snap,
                                                       vol=oldest))
            if len(hunged_vols) > 1:
                logger.warn(
                    'Replication to temporary {vols} created during '
                    'transmitting {snap} to {reg} qualified as hunged up. '
                    'Starting new replication process.'.format(
                        snap=src_snap, vols=hunged_vols, reg=dst_vol.region))
        update_snap(src_vol, src_mnt, dst_vol, dst_mnt, encr)

    if vol_snaps:
        dst_snap = sorted(vol_snaps, key=get_snap_time)[-1]
        with nested(
                attach_snapshot(src_snap, inst=src_inst, encr=encr),
                attach_snapshot(dst_snap, inst=dst_inst, encr=encr)) as (
                    (src_vol, src_mnt), (dst_vol, dst_mnt)):
            sync_mountpoints(src_snap, src_vol, src_mnt, dst_vol, dst_mnt)
    else:
        with nested(
                attach_snapshot(src_snap, inst=src_inst, encr=encr),
                create_tmp_volume(dst_conn.region, src_snap.volume_size)) as (
                    (src_vol, src_mnt), (dst_vol, dst_mnt)):
            sync_mountpoints(src_snap, src_vol, src_mnt, dst_vol, dst_mnt)


@task
def rsync_region(
        src_region_name, dst_region_name, tag_name=DEFAULT_TAG_NAME,
        tag_value=DEFAULT_TAG_VALUE, native_only=True):
    """Duplicates latest snapshots with given tag into dst_region.

    src_region_name, dst_region_name
        every latest volume snapshot from src_region will be rsynced
        to the dst_region;
    tag_name, tag_value
        snapshots will be filtered by tag. Tag will be fetched from
        config by default;
    native_only
        sync only snapshots, created in the src_region_name. True by
        default."""
    src_conn = get_region_conn(src_region_name)
    dst_conn = get_region_conn(dst_region_name)
    snaps = get_relevant_snapshots(src_conn, tag_name, tag_value, native_only)
    if not snaps:
        return
    with nested(create_temp_inst(src_conn.region),
                create_temp_inst(dst_conn.region)) as (src_inst, dst_inst):
        snaps = sorted(snaps, key=get_snap_vol)    # Prepare for grouping.
        for vol, vol_snaps in groupby(snaps, get_snap_vol):
            latest_snap = sorted(vol_snaps, key=get_snap_time)[-1]
            for inst in src_inst, dst_inst:
                logger.debug('Rebooting {0} in {0.region} '
                             'to refresh attachments'.format(inst))
                inst.reboot()
            args = (src_region_name, latest_snap.id, dst_region_name, src_inst,
                    dst_inst)
            try:
                rsync_snapshot(*args)
            except:
                logger.exception('rsync of {1} from {0} to {2} failed'.format(
                    *args))


@task
def rsync_all_regions(primary_backup_region, secondary_backup_region):
    """
    Replicates snapshots across all regions.

    Sync snapshots from all but primary regions into primary. And then
    snapshots from primary regions will be replicated into secondary
    backup region.

    :param primary_backup_region: AWS region name that keeps all
        snapshots clones;
    :type primary_backup_region: str
    :param secondary_backup_region: AWS region name that keeps clones of
        snapshots from `primary_backup_region`.
    :type secondary_backup_region: str
    """
    pri_name = get_region_conn(primary_backup_region).region.name
    all_regs = get_region_conn().get_all_regions()
    for reg in (reg for reg in all_regs if reg.name != pri_name):
        rsync_region(reg.name, primary_backup_region)
    rsync_region(primary_backup_region, secondary_backup_region)
